// CSC 2012
// Virtualization excercises

#include <sys/stat.h>

#include <unistd.h>
#include <cstdio>
#include <cstdlib>
#include <cstring>
#include <string>
#include <map>
#include "Snap.h"  // Standford Network Analysis Project (Graph library)


// U Brandes, "A faster algorithm for betweenness centrality",
// Journal of Mathematical Sociology (25), 163–177, 2001
static void Dependency(const PUNGraph &Graph, const TInt &RootNode,
                       TIntFltH &DepH)
{
  TSnapQueue<TInt> Q;  // breadth first search queue
  TSStack<TInt> S;  // vertices in order of distance from root vertex
  THash<TInt, TInt> BFSLevel;  // BFS ordering

  THash<TInt, TInt> Sigma;  // number of shortest paths
  THash<TInt, TIntV> P;  // direct predecessors on shortest paths

  // Initialization
  for (TUNGraph::TNodeI NI = Graph->BegNI(); NI < Graph->EndNI(); NI++) {
    BFSLevel.AddDat(NI.GetId()) = -1;
    Sigma.AddDat(NI.GetId()) = 0;
    DepH.AddDat(NI.GetId()) = 0.0;
  }
  BFSLevel.AddDat(RootNode) = 0;
  Sigma.AddDat(RootNode) = 1;
  P.AddDat(RootNode) = TIntV();
  Q.Push(RootNode);

  // BFS
  while (!Q.Empty()) {
    TInt vId = Q.Top();
    TUNGraph::TNodeI vNI = Graph->GetNI(vId);
    Q.Pop();
    S.Push(vId);

    for (int e = 0; e < vNI.GetOutDeg(); e++) {
      TInt wId = vNI.GetOutNId(e);
      if (BFSLevel.GetDat(wId) == -1) {
        // tree edge
        Q.Push(wId);
        BFSLevel.AddDat(wId) = BFSLevel.GetDat(vId) + 1;
      }
      if (BFSLevel.GetDat(wId) == BFSLevel.GetDat(vId) + 1) {
        // forward edge
        P.AddDat(wId).Add(vId);
        Sigma.AddDat(wId) = Sigma.GetDat(wId) + Sigma.GetDat(vId);
      }
    }  // all outgoing edges
  }  // node stack

  // finalize
  while (!S.Empty()) {
    TInt wId = S.Top();
    S.Pop();

    TIntV *wPredecessors = &P.GetDat(wId);
    for (TIntV::TIter vId = wPredecessors->BegI(); vId < wPredecessors->EndI();
         vId++)
    {
      DepH.AddDat(*vId) = DepH.AddDat(*vId) +
                            ((float)Sigma.GetDat(*vId)/(float)Sigma.GetDat(wId))
                            * (1.0 + DepH.GetDat(wId));
    }
  }
  DepH.AddDat(RootNode) = 0.0;
}


static void Usage(const std::string &bin_image) {
  printf("Computes the betweeness dependency of vertices in a graph.\n"
         "Usage:\n"
         "  %s -g <graph file> -s <first vertex> -e <last vertex>\n\n"
         "Example:\n"
         "  %s -g power-grid.graph -s 5 -e 7\n"
         "The output will be lines in the following format\n"
         "  <node id>  <partial betweeness>\n"
         "representing the dependency (the partial betweeness) of the vertices\n"
         "in the vertex range (here: 5--7) from all the other vertices.\n"
         "So in the output, all the vertex IDs will appear.\n",
         bin_image.c_str(), bin_image.c_str());
}


int main(int argc, char **argv) {
  std::string graph_path = "";
  int range_start = -1;
  int range_end = -1;

  // Option parsing
  if (argc < 2) {
    Usage(argv[0]);
    return 1;
  }

	char c;
	while ((c = getopt(argc, argv, "g:s:e:h")) != -1) {
		switch (c) {
      case 'g':
        graph_path = optarg;
        break;
      case 's':
        range_start = atoi(optarg);
        break;
      case 'e':
        range_end = atoi(optarg);
        break;
      case 'h':
      case '?':
      default:
        Usage(argv[0]);
        return 0;
    }
  }

  if ((graph_path == "") || (range_end < range_start) || (range_end < 0)) {
    Usage(argv[0]);
    return 1;
  }

  struct stat info;
  int retval = stat(graph_path.c_str(), &info);
  if (retval != 0) {
    fprintf(stderr, "failed to find input graph.\n");
    return 1;
  }
  PUNGraph Graph = TSnap::LoadEdgeList<PUNGraph>(graph_path.c_str());
  //fprintf(stderr, "Loaded graph has %d vertices and %d edges\n",
  //          Graph->GetNodes(), Graph->GetEdges());

  std::map<int, float> results;
  int i = 0;
  for (TUNGraph::TNodeI NI = Graph->BegNI(); NI < Graph->EndNI(); NI++, i++) {
    if (i < range_start)
      continue;
    if (i > range_end)
      break;

    TIntFltH DepH;
    Dependency(Graph, NI.GetId(), DepH);

    for (TUNGraph::TNodeI NJ = Graph->BegNI(); NJ < Graph->EndNI(); NJ++) {
      float partial_sum = DepH.GetDat(NJ.GetId())()/2.0;
      if (results.find(NJ.GetId()) == results.end()) {
        results[NJ.GetId()] = partial_sum;
      } else {
        results[NJ.GetId()] += partial_sum;
      }
    }
  }

  for (std::map<int, float>::const_iterator i = results.begin(),
       iEnd = results.end(); i != iEnd; ++i)
  {
    printf("%d %f\n", i->first, i->second);
  }

  return 0;
}
